@@ -34,6 +34,7 @@
 #include "yelp-error.h"
 #include "yelp-mallard-document.h"
 #include "yelp-settings.h"
+#include "yelp-storage.h"
 #include "yelp-transform.h"
 #include "yelp-debug.h"
 
@@ -429,6 +430,11 @@ mallard_try_run (YelpMallardDocument *mallard,
         return;
     }
 
+    if (page_data->transform != NULL) {
+        /* It's already running. Just let it be. */
+        return;
+    }
+
     mallard_page_data_run (page_data);
 }
 
